/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.anfrage;

import java.io.Serializable;
import java.util.HashMap;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.puk.param.hierarchie.HierarchieObjekt;
import de.bsvrz.puk.param.lib.Parameter;
import de.bsvrz.puk.param.lib.ParameterInfo;
import de.bsvrz.puk.param.lib.ParameterInfo.SerializableParameterInfo;
import de.bsvrz.puk.param.lib.ParametrierungsAnfrageTyp;
import de.bsvrz.puk.param.lib.SerializableParameter;
import de.bsvrz.puk.param.param.ParamMessage;
import de.bsvrz.puk.param.param.Parametrierung;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Realisierung einer Klasse zum Beantworten einer Anfrage zum Schreiben von
 * Parametern.<br>
 * Die Schnittstelle erlaubt zusätzlich zur Möglichkeit, Parameterdatensätze
 * über den Aspekt "asp.parameterVorgabe" und das
 * <code>ClientSenderInterface</code> hinaus, Parameterdatensätze an
 * Hierrchieobjekte und Typobjekte zu speichern, um sie mehreren untergeordneten
 * Systemobjekten als Daten zur Verfügung zu stellen.<br>
 * Außerdem bietet diese Schnittstelle die einzige Möglichkeit Parameter zu
 * löschen ohne den Parametrierungsdatensatz selbst zu manipulieren.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class ParameterSchreibenErmittler extends ParameterAnfrageBeantworter {

	/** Logger für Debugausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/** die Parameter die gesetzt werden sollen. */
	private final Parameter[] setzListe;

	/**
	 * Konstruktor, erzeugt eine Instanz des Ermittlers.<br>
	 * Die übergebenen Parameterdatensätze werden an das übergebene
	 * Parametrierungsmodul übergeben, das für die Speicherung und
	 * Veröffentlichung sorgt. An das Anfrageobjekt, welches ein Systemobjekt
	 * vom Typ "Appliktation" sein sollte, wird eine leere Antwort versendet,
	 * wenn alle Parameterdaten gesetzt wurden. Der Index wird in der Antwort
	 * zur Zuordnung von Antwort und Anfrage auf der Seote des anfragenden
	 * Clients eingefügt. Als Daten zum Erzeugen des Anfragebeantworters wird
	 * ein Array mit serialisierten Parameterdatensätzen erwartet.
	 *
	 * @param parametrierung
	 *            das verwendete Parametrierungsmodul
	 * @param anfrager
	 *            das Systemobjekt, das die anfragende Applikation repräsentiert
	 * @param index
	 *            der Index der Anfrage
	 * @param daten
	 *            die Parameterdaten der Anfrage
	 */
	ParameterSchreibenErmittler(final Parametrierung parametrierung, final SystemObject anfrager, final int index,
			final Object daten) {
		super(parametrierung, anfrager, index);

		if (!(daten instanceof SerializableParameter[])) {
			throw new IllegalArgumentException("Die Daten der Anfrage konnten nicht interpretiert werden: " + daten);
		}

		final SerializableParameter[] anfrageFeld = (SerializableParameter[]) daten;
		setzListe = new Parameter[anfrageFeld.length];

		for (int idx = 0; idx < anfrageFeld.length; idx++) {
			setzListe[idx] = anfrageFeld[idx].erzeugeParameter(parametrierung.getVerbindung());
		}
	}

	/**
	 * Es werden alle inerhalb der Anfrage übermittelten Parameterdatensätze
	 * innerhalb des Parametrierungsmoduls gespeichert und veröffentlicht. Die
	 * Antwort enthält keine Daten.
	 */
	@Override
	protected Serializable ermittleAntwort() {

		ParameterSchreibenErmittler.LOGGER.fine("Parameter schreiben für " + getAnfrager());

		final HashMap<SerializableParameterInfo, String> errors = new HashMap<>();

		for (final Parameter parameter : setzListe) {
			final ParameterInfo info = parameter.getInfo();
			if (getParametrierung().getProvider().isProvided(info)) {

				if (parameter.getObjekt().isOfType(parameter.getTyp())) {
					setzeLokalenParameter(parameter);
				} else if (parameter.getObjekt() instanceof SystemObjectType) {
					if (parameter.getTyp().equals(parameter.getObjekt())) {
						setzeTypParameter(parameter.getTyp(), parameter);
					} else {
						errors.put(new SerializableParameterInfo(info, getParametrierung().getVerbindung()),
								"Ungültige Parameterinformation");
						LogTools.log(ParameterSchreibenErmittler.LOGGER, ParamMessage.WarningFalscheTypKombination,
								parameter.getObjekt(), parameter.getTyp());
					}
				} else {
					final HierarchieObjekt hierarchieObjekt = getParametrierung().getHierarchieManager()
							.getHierarchieObjekt(parameter.getObjekt().getId());
					if (hierarchieObjekt != null) {
						if (hierarchieObjekt.isTypGueltig(parameter.getTyp())) {
							setzeHierarchieParameter(hierarchieObjekt, parameter);
						}
					}
				}
			} else {
				errors.put(new SerializableParameterInfo(info, getParametrierung().getVerbindung()),
						"Parameter wird nicht bereitgestellt");
			}
		}

		ParameterSchreibenErmittler.LOGGER.fine("Es wurden " + setzListe.length + " Parameterdatensätze geschrieben",
				setzListe.length);

		return errors;
	}

	@Override
	protected ParametrierungsAnfrageTyp getAntwortTyp() {
		return ParametrierungsAnfrageTyp.PARAMETER;
	}

	/**
	 * die Funktion setzt die Parameter eines Hierarchieobjekts und ruft die
	 * Funktion zum Publizieren der Parameter auf. Damit werden die Parameter
	 * aller dem Hierarchieobjekt untergeodneten Systemobjekte als
	 * Soll-Parameterdatensätze neu publiziert. Der Datensatz des
	 * Hierarchieobjekts selbst wird nur intern gespeichert und kann über die
	 * Parametrierungsschnittstelle ausgelesen werden.
	 *
	 * @param hierarchieObjekt
	 *            das Hierarchieobjekt
	 * @param parameter
	 *            die zu setzenden Parameter
	 */
	private void setzeHierarchieParameter(final HierarchieObjekt hierarchieObjekt, final Parameter parameter) {
		ParameterSchreibenErmittler.LOGGER.fine("Setze Hierarchieparameter für Objekt:",
				new Object[] { hierarchieObjekt, parameter });
		getParametrierung().getProvider().publiziereHierarchieParameter(hierarchieObjekt, parameter);
	}

	/**
	 * setzt die Daten eines lokalen Parameter. Die Daten werden dem
	 * Parametrierungsmodul übergeben und veröffentlicht.
	 *
	 * @param parameter
	 *            die Parameterdaten
	 */
	private void setzeLokalenParameter(final Parameter parameter) {
		ParameterSchreibenErmittler.LOGGER.fine("Setze lokalen Parameter" + parameter.getObjekt(), parameter);
		getParametrierung().getProvider().publiziereLokalenParameter(parameter);
	}

	/**
	 * setzt die Parameter eines Typobjekts. Die Daten werden an das
	 * Parametrierungsmodul übergeben und gespeichert, gleichzeitig wird die
	 * Publizierungsfunktion des Parameterproviders für das Typobjekt
	 * aufgerufen, womit die antsprechenden Parameterdaten aller betroffenen
	 * Objekte des jeweiligen Typs neu als Soll-Parameterdatensatz publiziert
	 * werden. Die Daten des Typobjekts selbst werden nicht publiziert und
	 * können über die Parametrierungsschnittstelle abgerufen werden.
	 *
	 * @param typ
	 *            der Typ an dem die Parameter gesetzt werden sollen
	 * @param parameter
	 *            die Parameterdaten
	 */
	private void setzeTypParameter(final SystemObjectType typ, final Parameter parameter) {
		ParameterSchreibenErmittler.LOGGER.fine("Setze Typparameter für den Typ: " + typ, parameter);
		getParametrierung().getProvider().publiziereTypParameter(typ, parameter);
	}
}
