/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.hierarchie;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Repräsentation der vom Nutzer im Datenkatalog der Konfiguration definierten
 * Hierarchieobjekt-Struktur.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class TypHierarchie {

	/** Logger für Debugausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/** die Menge der definierten Hierarchieobjekte. */
	private final Map<SystemObjectType, Set<TypHierarchieObjekt>> objekte = new HashMap<>();

	/**
	 * erzeugt die Typhierarchie für die gegebene Datenverteilerverbindung.<br>
	 * Es werden die Datensätze aller Objekte vom Typ "typ.hierarchieDefinition"
	 * ausgewertet und die entsprechenden Typhierarchieobjekte angelegt.
	 *
	 * @param verbindung
	 *            die verwendete Datenverteilerverbindung
	 */
	TypHierarchie(final ClientDavInterface verbindung) {

		final DataModel model = verbindung.getDataModel();

		for (final SystemObject object : verbindung.getDataModel().getType("typ.hierarchieDefinition").getElements()) {
			final Data data = object
					.getConfigurationData(verbindung.getDataModel().getAttributeGroup("atg.hierarchieDefinition"));
			final Data.Array objektListe = data.getArray("HierarchieObjekte");
			for (int idx = 0; idx < objektListe.getLength(); idx++) {
				final Data item = objektListe.getItem(idx);
				final SystemObjectType type = model.getType(item.getTextValue("ObjektTyp").getText());
				final String mengenName = item.getTextValue("Menge").getText();
				final SystemObjectType nachfolgerType = model.getType(item.getTextValue("NachfolgerTyp").getText());

				Set<TypHierarchieObjekt> set = objekte.get(type);
				if (set == null) {
					set = new HashSet<>();
					set.add(new TypHierarchieObjekt(type, mengenName, nachfolgerType));
					objekte.put(type, set);
				} else {
					set.add(new TypHierarchieObjekt(type, mengenName, nachfolgerType));
				}
			}
		}

		pruefeHierarchie();
		TypHierarchie.LOGGER.config("Typhierarchie initialisiert und geprüft");
	}

	/**
	 * ersetzt eine Liste von Typen durch ihre Super- oder Subtypen.
	 *
	 * @param suchListe
	 *            die Liste der Typen
	 * @param superTyp
	 *            Supertypen ermitteln ?
	 * @return die erweiterte Liste
	 */
	private List<SystemObjectType> erweitereTypListe(final List<SystemObjectType> suchListe, final boolean superTyp) {
		final List<SystemObjectType> result = new ArrayList<>();
		for (final SystemObjectType type : suchListe) {
			if (superTyp) {
				result.addAll(type.getSuperTypes());
			} else {
				result.addAll(type.getSubTypes());
			}
		}
		return result;
	}

	/**
	 * liefert die Menge aller Typhierarchieeinträge.
	 *
	 * @return die Menge der Einträge
	 */
	Collection<TypHierarchieObjekt> getEintraege() {
		final Collection<TypHierarchieObjekt> result = new ArrayList<>();
		for (final Set<TypHierarchieObjekt> set : objekte.values()) {
			result.addAll(set);
		}
		return result;
	}

	/**
	 * prüft, ob die ermittelte Typhierarchie zyklenfrei ist.
	 */
	private void pruefeHierarchie() {
		for (final Set<TypHierarchieObjekt> set : objekte.values()) {
			for (final TypHierarchieObjekt objekt : set) {
				final Set<SystemObjectType> benutzt = new HashSet<>();
				final List<TypHierarchieObjekt> suchListe = new ArrayList<>();
				suchListe.add(objekt);

				while (suchListe.size() > 0) {
					final TypHierarchieObjekt suchObjekt = suchListe.get(0);
					suchListe.remove(0);
					benutzt.add(suchObjekt.getTyp());

					final SystemObjectType nachfolgerTyp = suchObjekt.getNachfolgerTyp();
					if (nachfolgerTyp != null) {

						SystemObjectType verwendet = null;
						if (benutzt.contains(nachfolgerTyp)) {
							verwendet = nachfolgerTyp;
						} else {
							verwendet = pruefeTypListe(nachfolgerTyp, benutzt, false);
							if (verwendet == null) {
								verwendet = pruefeTypListe(nachfolgerTyp, benutzt, true);
							}
						}

						if (verwendet != null) {
							throw new IllegalStateException("Die definierte Typhierarchie ist nicht zyklenfrei: "
									+ nachfolgerTyp + " wird auf innerhalb der Hierarchie von " + objekt.getTyp()
									+ " bereits als " + verwendet + " verwendet");
						}

						final Set<TypHierarchieObjekt> nachfolgerObjekte = objekte.get(nachfolgerTyp);
						if (nachfolgerObjekte != null) {
							suchListe.addAll(nachfolgerObjekte);
						}
					}
				}
			}
		}
	}

	/**
	 * ermittelt ob der übergebene Typ oder einer seiner Sub- oder Supertypen
	 * bereits in der Hierarchie verwendet wird.
	 *
	 * @param neuerTyp
	 *            der zu prüfende Typ
	 * @param benutzt
	 *            die Menge der bereits verwendeten Typen
	 * @param superTyp
	 *            Supertypen bewerten ?
	 * @return das Typobjekt des bereits verwendeten Typs
	 */
	private SystemObjectType pruefeTypListe(final SystemObjectType neuerTyp, final Set<SystemObjectType> benutzt,
			final boolean superTyp) {
		SystemObjectType result = null;

		List<SystemObjectType> suchListe = null;
		if (superTyp) {
			suchListe = neuerTyp.getSuperTypes();
		} else {
			suchListe = neuerTyp.getSubTypes();
		}
		while (suchListe.size() > 0) {
			for (final SystemObjectType type : suchListe) {
				if (benutzt.contains(type)) {
					result = type;
					break;
				}
			}

			if (result != null) {
				break;
			}
			suchListe = erweitereTypListe(suchListe, superTyp);
		}

		return result;
	}
}
